#!/usr/bin/env python

import hashlib
import os
import os.path
import shutil
import struct
import subprocess
import sys
import zipfile
from google.protobuf import message

# from https://android.googlesource.com/platform/system/update_engine/+/refs/heads/master/scripts/update_payload/
from . import update_metadata_pb2

PROGRAMS = [ 'bzcat', 'xzcat' ]

BRILLO_MAJOR_PAYLOAD_VERSION = 2

class PayloadError(Exception):
  pass

class Payload(object):
  class _PayloadHeader(object):
    _MAGIC = b'CrAU'

    def __init__(self):
      self.version = None
      self.manifest_len = None
      self.metadata_signature_len = None
      self.size = None

    def ReadFromPayload(self, payload_file):
      magic = payload_file.read(4)
      if magic != self._MAGIC:
        raise PayloadError('Invalid payload magic: %s' % magic)
      self.version = struct.unpack('>Q', payload_file.read(8))[0]
      self.manifest_len = struct.unpack('>Q', payload_file.read(8))[0]
      self.size = 20
      self.metadata_signature_len = 0
      if self.version != BRILLO_MAJOR_PAYLOAD_VERSION:
        raise PayloadError('Unsupported payload version (%d)' % self.version)
      self.size += 4
      self.metadata_signature_len = struct.unpack('>I', payload_file.read(4))[0]

  def __init__(self, payload_file):
    self.payload_file = payload_file
    self.header = None
    self.manifest = None
    self.data_offset = None
    self.metadata_signature = None
    self.metadata_size = None

  def _ReadManifest(self):
    return self.payload_file.read(self.header.manifest_len)

  def _ReadMetadataSignature(self):
    self.payload_file.seek(self.header.size + self.header.manifest_len)
    return self.payload_file.read(self.header.metadata_signature_len);

  def ReadDataBlob(self, offset, length):
    self.payload_file.seek(self.data_offset + offset)
    return self.payload_file.read(length)

  def Init(self):
    self.header = self._PayloadHeader()
    self.header.ReadFromPayload(self.payload_file)
    manifest_raw = self._ReadManifest()
    self.manifest = update_metadata_pb2.DeltaArchiveManifest()
    try:
        self.manifest.ParseFromString(manifest_raw)
    except message.DecodeError as parse_error:
        print("[WARN] Cannot deserialize Protobuf. Reason: {}".format(parse_error))
    metadata_signature_raw = self._ReadMetadataSignature()
    if metadata_signature_raw:
      self.metadata_signature = update_metadata_pb2.Signatures()
      self.metadata_signature.ParseFromString(metadata_signature_raw)
    self.metadata_size = self.header.size + self.header.manifest_len
    self.data_offset = self.metadata_size + self.header.metadata_signature_len

def decompress_payload(command, data, size, hash):
  p = subprocess.Popen([command, '-'], stdout=subprocess.PIPE, stdin=subprocess.PIPE)
  r = p.communicate(data)[0]
  if len(r) != size:
    print("Unexpected size %d %d" % (len(r), size))
  elif hashlib.sha256(data).digest() != hash:
    print("Hash mismatch")
  return r

def parse_payload(payload_f, partition, out_f):
  BLOCK_SIZE = 4096
  for operation in partition.operations:
    e = operation.dst_extents[0]
    data = payload_f.ReadDataBlob(operation.data_offset, operation.data_length)
    out_f.seek(e.start_block * BLOCK_SIZE)
    if operation.type == update_metadata_pb2.InstallOperation.REPLACE:
      out_f.write(data)
    elif operation.type == update_metadata_pb2.InstallOperation.REPLACE_XZ:
      r = decompress_payload('xzcat', data, e.num_blocks * BLOCK_SIZE, operation.data_sha256_hash)
      out_f.write(r)
    elif operation.type == update_metadata_pb2.InstallOperation.REPLACE_BZ:
      r = decompress_payload('bzcat', data, e.num_blocks * BLOCK_SIZE, operation.data_sha256_hash)
      out_f.write(r)
    elif operation.type == update_metadata_pb2.InstallOperation.ZERO:
      out_f.write(b'\x00' * (e.num_blocks * BLOCK_SIZE))
    else:
      raise PayloadError('Unhandled operation type ({} - {})'.format(operation.type,
                         update_metadata_pb2.InstallOperation.Type.Name(operation.type)))

def main(filename, output_dir):
  if filename.endswith('.zip'):
    print("Extracting 'payload.bin' from OTA file...")
    ota_zf = zipfile.ZipFile(filename)
    payload_file = open(ota_zf.extract('payload.bin', output_dir), 'rb')
  else:
    payload_file = open(filename, 'rb')

  payload = Payload(payload_file)
  payload.Init()

  for p in payload.manifest.partitions:
    name = p.partition_name + '.img'
    print("Extracting '%s'" % name)
    fname = os.path.join(output_dir, name)
    out_f = open(fname, 'wb')
    try:
      parse_payload(payload, p, out_f)
    except PayloadError as e:
      print('Failed: %s' % e)
      out_f.close()
      os.unlink(fname)

if __name__ == '__main__':
  try:
    filename = sys.argv[1]
  except:
    print('Usage: %s payload.bin [output_dir]' % sys.argv[0])
    sys.exit()

  try:
    output_dir = sys.argv[2]
  except IndexError:
    output_dir = os.getcwd()

  if not os.path.exists(output_dir):
    os.makedirs(output_dir)

  main(filename, output_dir)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: update_metadata.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='update_metadata.proto',
  package='chromeos_update_engine',
  syntax='proto2',
  serialized_options=_b('H\003'),
  serialized_pb=_b('\n\x15update_metadata.proto\x12\x16\x63hromeos_update_engine\"1\n\x06\x45xtent\x12\x13\n\x0bstart_block\x18\x01 \x01(\x04\x12\x12\n\nnum_blocks\x18\x02 \x01(\x04\"\x9f\x01\n\nSignatures\x12@\n\nsignatures\x18\x01 \x03(\x0b\x32,.chromeos_update_engine.Signatures.Signature\x1aO\n\tSignature\x12\x13\n\x07version\x18\x01 \x01(\rB\x02\x18\x01\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x1f\n\x17unpadded_signature_size\x18\x03 \x01(\x07\"+\n\rPartitionInfo\x12\x0c\n\x04size\x18\x01 \x01(\x04\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"\x8f\x01\n\tImageInfo\x12\x11\n\x05\x62oard\x18\x01 \x01(\tB\x02\x18\x01\x12\x0f\n\x03key\x18\x02 \x01(\tB\x02\x18\x01\x12\x13\n\x07\x63hannel\x18\x03 \x01(\tB\x02\x18\x01\x12\x13\n\x07version\x18\x04 \x01(\tB\x02\x18\x01\x12\x19\n\rbuild_channel\x18\x05 \x01(\tB\x02\x18\x01\x12\x19\n\rbuild_version\x18\x06 \x01(\tB\x02\x18\x01\"\xfc\x03\n\x10InstallOperation\x12;\n\x04type\x18\x01 \x02(\x0e\x32-.chromeos_update_engine.InstallOperation.Type\x12\x13\n\x0b\x64\x61ta_offset\x18\x02 \x01(\x04\x12\x13\n\x0b\x64\x61ta_length\x18\x03 \x01(\x04\x12\x33\n\x0bsrc_extents\x18\x04 \x03(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x12\n\nsrc_length\x18\x05 \x01(\x04\x12\x33\n\x0b\x64st_extents\x18\x06 \x03(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x12\n\ndst_length\x18\x07 \x01(\x04\x12\x18\n\x10\x64\x61ta_sha256_hash\x18\x08 \x01(\x0c\x12\x17\n\x0fsrc_sha256_hash\x18\t \x01(\x0c\"\xbb\x01\n\x04Type\x12\x0b\n\x07REPLACE\x10\x00\x12\x0e\n\nREPLACE_BZ\x10\x01\x12\x0c\n\x04MOVE\x10\x02\x1a\x02\x08\x01\x12\x0e\n\x06\x42SDIFF\x10\x03\x1a\x02\x08\x01\x12\x0f\n\x0bSOURCE_COPY\x10\x04\x12\x11\n\rSOURCE_BSDIFF\x10\x05\x12\x0e\n\nREPLACE_XZ\x10\x08\x12\x08\n\x04ZERO\x10\x06\x12\x0b\n\x07\x44ISCARD\x10\x07\x12\x11\n\rBROTLI_BSDIFF\x10\n\x12\x0c\n\x08PUFFDIFF\x10\t\x12\x0c\n\x08ZUCCHINI\x10\x0b\"\x81\x02\n\x11\x43owMergeOperation\x12<\n\x04type\x18\x01 \x01(\x0e\x32..chromeos_update_engine.CowMergeOperation.Type\x12\x32\n\nsrc_extent\x18\x02 \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x32\n\ndst_extent\x18\x03 \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x12\n\nsrc_offset\x18\x04 \x01(\r\"2\n\x04Type\x12\x0c\n\x08\x43OW_COPY\x10\x00\x12\x0b\n\x07\x43OW_XOR\x10\x01\x12\x0f\n\x0b\x43OW_REPLACE\x10\x02\"\xc8\x06\n\x0fPartitionUpdate\x12\x16\n\x0epartition_name\x18\x01 \x02(\t\x12\x17\n\x0frun_postinstall\x18\x02 \x01(\x08\x12\x18\n\x10postinstall_path\x18\x03 \x01(\t\x12\x17\n\x0f\x66ilesystem_type\x18\x04 \x01(\t\x12M\n\x17new_partition_signature\x18\x05 \x03(\x0b\x32,.chromeos_update_engine.Signatures.Signature\x12\x41\n\x12old_partition_info\x18\x06 \x01(\x0b\x32%.chromeos_update_engine.PartitionInfo\x12\x41\n\x12new_partition_info\x18\x07 \x01(\x0b\x32%.chromeos_update_engine.PartitionInfo\x12<\n\noperations\x18\x08 \x03(\x0b\x32(.chromeos_update_engine.InstallOperation\x12\x1c\n\x14postinstall_optional\x18\t \x01(\x08\x12=\n\x15hash_tree_data_extent\x18\n \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x38\n\x10hash_tree_extent\x18\x0b \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x1b\n\x13hash_tree_algorithm\x18\x0c \x01(\t\x12\x16\n\x0ehash_tree_salt\x18\r \x01(\x0c\x12\x37\n\x0f\x66\x65\x63_data_extent\x18\x0e \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x32\n\nfec_extent\x18\x0f \x01(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x14\n\tfec_roots\x18\x10 \x01(\r:\x01\x32\x12\x0f\n\x07version\x18\x11 \x01(\t\x12\x43\n\x10merge_operations\x18\x12 \x03(\x0b\x32).chromeos_update_engine.CowMergeOperation\x12\x19\n\x11\x65stimate_cow_size\x18\x13 \x01(\x04\"L\n\x15\x44ynamicPartitionGroup\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0c\n\x04size\x18\x02 \x01(\x04\x12\x17\n\x0fpartition_names\x18\x03 \x03(\t\"\xbe\x01\n\x18\x44ynamicPartitionMetadata\x12=\n\x06groups\x18\x01 \x03(\x0b\x32-.chromeos_update_engine.DynamicPartitionGroup\x12\x18\n\x10snapshot_enabled\x18\x02 \x01(\x08\x12\x14\n\x0cvabc_enabled\x18\x03 \x01(\x08\x12\x1e\n\x16vabc_compression_param\x18\x04 \x01(\t\x12\x13\n\x0b\x63ow_version\x18\x05 \x01(\r\"c\n\x08\x41pexInfo\x12\x14\n\x0cpackage_name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\x03\x12\x15\n\ris_compressed\x18\x03 \x01(\x08\x12\x19\n\x11\x64\x65\x63ompressed_size\x18\x04 \x01(\x03\"C\n\x0c\x41pexMetadata\x12\x33\n\tapex_info\x18\x01 \x03(\x0b\x32 .chromeos_update_engine.ApexInfo\"\x9e\x07\n\x14\x44\x65ltaArchiveManifest\x12H\n\x12install_operations\x18\x01 \x03(\x0b\x32(.chromeos_update_engine.InstallOperationB\x02\x18\x01\x12O\n\x19kernel_install_operations\x18\x02 \x03(\x0b\x32(.chromeos_update_engine.InstallOperationB\x02\x18\x01\x12\x18\n\nblock_size\x18\x03 \x01(\r:\x04\x34\x30\x39\x36\x12\x19\n\x11signatures_offset\x18\x04 \x01(\x04\x12\x17\n\x0fsignatures_size\x18\x05 \x01(\x04\x12\x42\n\x0fold_kernel_info\x18\x06 \x01(\x0b\x32%.chromeos_update_engine.PartitionInfoB\x02\x18\x01\x12\x42\n\x0fnew_kernel_info\x18\x07 \x01(\x0b\x32%.chromeos_update_engine.PartitionInfoB\x02\x18\x01\x12\x42\n\x0fold_rootfs_info\x18\x08 \x01(\x0b\x32%.chromeos_update_engine.PartitionInfoB\x02\x18\x01\x12\x42\n\x0fnew_rootfs_info\x18\t \x01(\x0b\x32%.chromeos_update_engine.PartitionInfoB\x02\x18\x01\x12=\n\x0eold_image_info\x18\n \x01(\x0b\x32!.chromeos_update_engine.ImageInfoB\x02\x18\x01\x12=\n\x0enew_image_info\x18\x0b \x01(\x0b\x32!.chromeos_update_engine.ImageInfoB\x02\x18\x01\x12\x18\n\rminor_version\x18\x0c \x01(\r:\x01\x30\x12;\n\npartitions\x18\r \x03(\x0b\x32\'.chromeos_update_engine.PartitionUpdate\x12\x15\n\rmax_timestamp\x18\x0e \x01(\x03\x12T\n\x1a\x64ynamic_partition_metadata\x18\x0f \x01(\x0b\x32\x30.chromeos_update_engine.DynamicPartitionMetadata\x12\x16\n\x0epartial_update\x18\x10 \x01(\x08\x12\x33\n\tapex_info\x18\x11 \x03(\x0b\x32 .chromeos_update_engine.ApexInfoB\x02H\x03')
)



_INSTALLOPERATION_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='chromeos_update_engine.InstallOperation.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REPLACE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPLACE_BZ', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOVE', index=2, number=2,
      serialized_options=_b('\010\001'),
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BSDIFF', index=3, number=3,
      serialized_options=_b('\010\001'),
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOURCE_COPY', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOURCE_BSDIFF', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPLACE_XZ', index=6, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ZERO', index=7, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISCARD', index=8, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BROTLI_BSDIFF', index=9, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PUFFDIFF', index=10, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ZUCCHINI', index=11, number=11,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=775,
  serialized_end=962,
)
_sym_db.RegisterEnumDescriptor(_INSTALLOPERATION_TYPE)

_COWMERGEOPERATION_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='chromeos_update_engine.CowMergeOperation.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COW_COPY', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COW_XOR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COW_REPLACE', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1172,
  serialized_end=1222,
)
_sym_db.RegisterEnumDescriptor(_COWMERGEOPERATION_TYPE)


_EXTENT = _descriptor.Descriptor(
  name='Extent',
  full_name='chromeos_update_engine.Extent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_block', full_name='chromeos_update_engine.Extent.start_block', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_blocks', full_name='chromeos_update_engine.Extent.num_blocks', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=98,
)


_SIGNATURES_SIGNATURE = _descriptor.Descriptor(
  name='Signature',
  full_name='chromeos_update_engine.Signatures.Signature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='chromeos_update_engine.Signatures.Signature.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='chromeos_update_engine.Signatures.Signature.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unpadded_signature_size', full_name='chromeos_update_engine.Signatures.Signature.unpadded_signature_size', index=2,
      number=3, type=7, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=181,
  serialized_end=260,
)

_SIGNATURES = _descriptor.Descriptor(
  name='Signatures',
  full_name='chromeos_update_engine.Signatures',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='signatures', full_name='chromeos_update_engine.Signatures.signatures', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SIGNATURES_SIGNATURE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=260,
)


_PARTITIONINFO = _descriptor.Descriptor(
  name='PartitionInfo',
  full_name='chromeos_update_engine.PartitionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='size', full_name='chromeos_update_engine.PartitionInfo.size', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash', full_name='chromeos_update_engine.PartitionInfo.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=305,
)


_IMAGEINFO = _descriptor.Descriptor(
  name='ImageInfo',
  full_name='chromeos_update_engine.ImageInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='board', full_name='chromeos_update_engine.ImageInfo.board', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='chromeos_update_engine.ImageInfo.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel', full_name='chromeos_update_engine.ImageInfo.channel', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='chromeos_update_engine.ImageInfo.version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='build_channel', full_name='chromeos_update_engine.ImageInfo.build_channel', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='build_version', full_name='chromeos_update_engine.ImageInfo.build_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=308,
  serialized_end=451,
)


_INSTALLOPERATION = _descriptor.Descriptor(
  name='InstallOperation',
  full_name='chromeos_update_engine.InstallOperation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='chromeos_update_engine.InstallOperation.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_offset', full_name='chromeos_update_engine.InstallOperation.data_offset', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_length', full_name='chromeos_update_engine.InstallOperation.data_length', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_extents', full_name='chromeos_update_engine.InstallOperation.src_extents', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_length', full_name='chromeos_update_engine.InstallOperation.src_length', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dst_extents', full_name='chromeos_update_engine.InstallOperation.dst_extents', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dst_length', full_name='chromeos_update_engine.InstallOperation.dst_length', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_sha256_hash', full_name='chromeos_update_engine.InstallOperation.data_sha256_hash', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_sha256_hash', full_name='chromeos_update_engine.InstallOperation.src_sha256_hash', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _INSTALLOPERATION_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=454,
  serialized_end=962,
)


_COWMERGEOPERATION = _descriptor.Descriptor(
  name='CowMergeOperation',
  full_name='chromeos_update_engine.CowMergeOperation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='chromeos_update_engine.CowMergeOperation.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_extent', full_name='chromeos_update_engine.CowMergeOperation.src_extent', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dst_extent', full_name='chromeos_update_engine.CowMergeOperation.dst_extent', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_offset', full_name='chromeos_update_engine.CowMergeOperation.src_offset', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _COWMERGEOPERATION_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=965,
  serialized_end=1222,
)


_PARTITIONUPDATE = _descriptor.Descriptor(
  name='PartitionUpdate',
  full_name='chromeos_update_engine.PartitionUpdate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='partition_name', full_name='chromeos_update_engine.PartitionUpdate.partition_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='run_postinstall', full_name='chromeos_update_engine.PartitionUpdate.run_postinstall', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='postinstall_path', full_name='chromeos_update_engine.PartitionUpdate.postinstall_path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filesystem_type', full_name='chromeos_update_engine.PartitionUpdate.filesystem_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_partition_signature', full_name='chromeos_update_engine.PartitionUpdate.new_partition_signature', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='old_partition_info', full_name='chromeos_update_engine.PartitionUpdate.old_partition_info', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_partition_info', full_name='chromeos_update_engine.PartitionUpdate.new_partition_info', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operations', full_name='chromeos_update_engine.PartitionUpdate.operations', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='postinstall_optional', full_name='chromeos_update_engine.PartitionUpdate.postinstall_optional', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_tree_data_extent', full_name='chromeos_update_engine.PartitionUpdate.hash_tree_data_extent', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_tree_extent', full_name='chromeos_update_engine.PartitionUpdate.hash_tree_extent', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_tree_algorithm', full_name='chromeos_update_engine.PartitionUpdate.hash_tree_algorithm', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash_tree_salt', full_name='chromeos_update_engine.PartitionUpdate.hash_tree_salt', index=12,
      number=13, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fec_data_extent', full_name='chromeos_update_engine.PartitionUpdate.fec_data_extent', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fec_extent', full_name='chromeos_update_engine.PartitionUpdate.fec_extent', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fec_roots', full_name='chromeos_update_engine.PartitionUpdate.fec_roots', index=15,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='chromeos_update_engine.PartitionUpdate.version', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='merge_operations', full_name='chromeos_update_engine.PartitionUpdate.merge_operations', index=17,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='estimate_cow_size', full_name='chromeos_update_engine.PartitionUpdate.estimate_cow_size', index=18,
      number=19, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1225,
  serialized_end=2065,
)


_DYNAMICPARTITIONGROUP = _descriptor.Descriptor(
  name='DynamicPartitionGroup',
  full_name='chromeos_update_engine.DynamicPartitionGroup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='chromeos_update_engine.DynamicPartitionGroup.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='chromeos_update_engine.DynamicPartitionGroup.size', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='partition_names', full_name='chromeos_update_engine.DynamicPartitionGroup.partition_names', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2067,
  serialized_end=2143,
)


_DYNAMICPARTITIONMETADATA = _descriptor.Descriptor(
  name='DynamicPartitionMetadata',
  full_name='chromeos_update_engine.DynamicPartitionMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='groups', full_name='chromeos_update_engine.DynamicPartitionMetadata.groups', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot_enabled', full_name='chromeos_update_engine.DynamicPartitionMetadata.snapshot_enabled', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vabc_enabled', full_name='chromeos_update_engine.DynamicPartitionMetadata.vabc_enabled', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vabc_compression_param', full_name='chromeos_update_engine.DynamicPartitionMetadata.vabc_compression_param', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cow_version', full_name='chromeos_update_engine.DynamicPartitionMetadata.cow_version', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2146,
  serialized_end=2336,
)


_APEXINFO = _descriptor.Descriptor(
  name='ApexInfo',
  full_name='chromeos_update_engine.ApexInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_name', full_name='chromeos_update_engine.ApexInfo.package_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='chromeos_update_engine.ApexInfo.version', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_compressed', full_name='chromeos_update_engine.ApexInfo.is_compressed', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decompressed_size', full_name='chromeos_update_engine.ApexInfo.decompressed_size', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2338,
  serialized_end=2437,
)


_APEXMETADATA = _descriptor.Descriptor(
  name='ApexMetadata',
  full_name='chromeos_update_engine.ApexMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='apex_info', full_name='chromeos_update_engine.ApexMetadata.apex_info', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2439,
  serialized_end=2506,
)


_DELTAARCHIVEMANIFEST = _descriptor.Descriptor(
  name='DeltaArchiveManifest',
  full_name='chromeos_update_engine.DeltaArchiveManifest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='install_operations', full_name='chromeos_update_engine.DeltaArchiveManifest.install_operations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_install_operations', full_name='chromeos_update_engine.DeltaArchiveManifest.kernel_install_operations', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_size', full_name='chromeos_update_engine.DeltaArchiveManifest.block_size', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=4096,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatures_offset', full_name='chromeos_update_engine.DeltaArchiveManifest.signatures_offset', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signatures_size', full_name='chromeos_update_engine.DeltaArchiveManifest.signatures_size', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='old_kernel_info', full_name='chromeos_update_engine.DeltaArchiveManifest.old_kernel_info', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_kernel_info', full_name='chromeos_update_engine.DeltaArchiveManifest.new_kernel_info', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='old_rootfs_info', full_name='chromeos_update_engine.DeltaArchiveManifest.old_rootfs_info', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_rootfs_info', full_name='chromeos_update_engine.DeltaArchiveManifest.new_rootfs_info', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='old_image_info', full_name='chromeos_update_engine.DeltaArchiveManifest.old_image_info', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_image_info', full_name='chromeos_update_engine.DeltaArchiveManifest.new_image_info', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='minor_version', full_name='chromeos_update_engine.DeltaArchiveManifest.minor_version', index=11,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='partitions', full_name='chromeos_update_engine.DeltaArchiveManifest.partitions', index=12,
      number=13, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_timestamp', full_name='chromeos_update_engine.DeltaArchiveManifest.max_timestamp', index=13,
      number=14, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dynamic_partition_metadata', full_name='chromeos_update_engine.DeltaArchiveManifest.dynamic_partition_metadata', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='partial_update', full_name='chromeos_update_engine.DeltaArchiveManifest.partial_update', index=15,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='apex_info', full_name='chromeos_update_engine.DeltaArchiveManifest.apex_info', index=16,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2509,
  serialized_end=3435,
)

_SIGNATURES_SIGNATURE.containing_type = _SIGNATURES
_SIGNATURES.fields_by_name['signatures'].message_type = _SIGNATURES_SIGNATURE
_INSTALLOPERATION.fields_by_name['type'].enum_type = _INSTALLOPERATION_TYPE
_INSTALLOPERATION.fields_by_name['src_extents'].message_type = _EXTENT
_INSTALLOPERATION.fields_by_name['dst_extents'].message_type = _EXTENT
_INSTALLOPERATION_TYPE.containing_type = _INSTALLOPERATION
_COWMERGEOPERATION.fields_by_name['type'].enum_type = _COWMERGEOPERATION_TYPE
_COWMERGEOPERATION.fields_by_name['src_extent'].message_type = _EXTENT
_COWMERGEOPERATION.fields_by_name['dst_extent'].message_type = _EXTENT
_COWMERGEOPERATION_TYPE.containing_type = _COWMERGEOPERATION
_PARTITIONUPDATE.fields_by_name['new_partition_signature'].message_type = _SIGNATURES_SIGNATURE
_PARTITIONUPDATE.fields_by_name['old_partition_info'].message_type = _PARTITIONINFO
_PARTITIONUPDATE.fields_by_name['new_partition_info'].message_type = _PARTITIONINFO
_PARTITIONUPDATE.fields_by_name['operations'].message_type = _INSTALLOPERATION
_PARTITIONUPDATE.fields_by_name['hash_tree_data_extent'].message_type = _EXTENT
_PARTITIONUPDATE.fields_by_name['hash_tree_extent'].message_type = _EXTENT
_PARTITIONUPDATE.fields_by_name['fec_data_extent'].message_type = _EXTENT
_PARTITIONUPDATE.fields_by_name['fec_extent'].message_type = _EXTENT
_PARTITIONUPDATE.fields_by_name['merge_operations'].message_type = _COWMERGEOPERATION
_DYNAMICPARTITIONMETADATA.fields_by_name['groups'].message_type = _DYNAMICPARTITIONGROUP
_APEXMETADATA.fields_by_name['apex_info'].message_type = _APEXINFO
_DELTAARCHIVEMANIFEST.fields_by_name['install_operations'].message_type = _INSTALLOPERATION
_DELTAARCHIVEMANIFEST.fields_by_name['kernel_install_operations'].message_type = _INSTALLOPERATION
_DELTAARCHIVEMANIFEST.fields_by_name['old_kernel_info'].message_type = _PARTITIONINFO
_DELTAARCHIVEMANIFEST.fields_by_name['new_kernel_info'].message_type = _PARTITIONINFO
_DELTAARCHIVEMANIFEST.fields_by_name['old_rootfs_info'].message_type = _PARTITIONINFO
_DELTAARCHIVEMANIFEST.fields_by_name['new_rootfs_info'].message_type = _PARTITIONINFO
_DELTAARCHIVEMANIFEST.fields_by_name['old_image_info'].message_type = _IMAGEINFO
_DELTAARCHIVEMANIFEST.fields_by_name['new_image_info'].message_type = _IMAGEINFO
_DELTAARCHIVEMANIFEST.fields_by_name['partitions'].message_type = _PARTITIONUPDATE
_DELTAARCHIVEMANIFEST.fields_by_name['dynamic_partition_metadata'].message_type = _DYNAMICPARTITIONMETADATA
_DELTAARCHIVEMANIFEST.fields_by_name['apex_info'].message_type = _APEXINFO
DESCRIPTOR.message_types_by_name['Extent'] = _EXTENT
DESCRIPTOR.message_types_by_name['Signatures'] = _SIGNATURES
DESCRIPTOR.message_types_by_name['PartitionInfo'] = _PARTITIONINFO
DESCRIPTOR.message_types_by_name['ImageInfo'] = _IMAGEINFO
DESCRIPTOR.message_types_by_name['InstallOperation'] = _INSTALLOPERATION
DESCRIPTOR.message_types_by_name['CowMergeOperation'] = _COWMERGEOPERATION
DESCRIPTOR.message_types_by_name['PartitionUpdate'] = _PARTITIONUPDATE
DESCRIPTOR.message_types_by_name['DynamicPartitionGroup'] = _DYNAMICPARTITIONGROUP
DESCRIPTOR.message_types_by_name['DynamicPartitionMetadata'] = _DYNAMICPARTITIONMETADATA
DESCRIPTOR.message_types_by_name['ApexInfo'] = _APEXINFO
DESCRIPTOR.message_types_by_name['ApexMetadata'] = _APEXMETADATA
DESCRIPTOR.message_types_by_name['DeltaArchiveManifest'] = _DELTAARCHIVEMANIFEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Extent = _reflection.GeneratedProtocolMessageType('Extent', (_message.Message,), {
  'DESCRIPTOR' : _EXTENT,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.Extent)
  })
_sym_db.RegisterMessage(Extent)

Signatures = _reflection.GeneratedProtocolMessageType('Signatures', (_message.Message,), {

  'Signature' : _reflection.GeneratedProtocolMessageType('Signature', (_message.Message,), {
    'DESCRIPTOR' : _SIGNATURES_SIGNATURE,
    '__module__' : 'update_metadata_pb2'
    # @@protoc_insertion_point(class_scope:chromeos_update_engine.Signatures.Signature)
    })
  ,
  'DESCRIPTOR' : _SIGNATURES,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.Signatures)
  })
_sym_db.RegisterMessage(Signatures)
_sym_db.RegisterMessage(Signatures.Signature)

PartitionInfo = _reflection.GeneratedProtocolMessageType('PartitionInfo', (_message.Message,), {
  'DESCRIPTOR' : _PARTITIONINFO,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.PartitionInfo)
  })
_sym_db.RegisterMessage(PartitionInfo)

ImageInfo = _reflection.GeneratedProtocolMessageType('ImageInfo', (_message.Message,), {
  'DESCRIPTOR' : _IMAGEINFO,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.ImageInfo)
  })
_sym_db.RegisterMessage(ImageInfo)

InstallOperation = _reflection.GeneratedProtocolMessageType('InstallOperation', (_message.Message,), {
  'DESCRIPTOR' : _INSTALLOPERATION,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.InstallOperation)
  })
_sym_db.RegisterMessage(InstallOperation)

CowMergeOperation = _reflection.GeneratedProtocolMessageType('CowMergeOperation', (_message.Message,), {
  'DESCRIPTOR' : _COWMERGEOPERATION,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.CowMergeOperation)
  })
_sym_db.RegisterMessage(CowMergeOperation)

PartitionUpdate = _reflection.GeneratedProtocolMessageType('PartitionUpdate', (_message.Message,), {
  'DESCRIPTOR' : _PARTITIONUPDATE,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.PartitionUpdate)
  })
_sym_db.RegisterMessage(PartitionUpdate)

DynamicPartitionGroup = _reflection.GeneratedProtocolMessageType('DynamicPartitionGroup', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMICPARTITIONGROUP,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.DynamicPartitionGroup)
  })
_sym_db.RegisterMessage(DynamicPartitionGroup)

DynamicPartitionMetadata = _reflection.GeneratedProtocolMessageType('DynamicPartitionMetadata', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMICPARTITIONMETADATA,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.DynamicPartitionMetadata)
  })
_sym_db.RegisterMessage(DynamicPartitionMetadata)

ApexInfo = _reflection.GeneratedProtocolMessageType('ApexInfo', (_message.Message,), {
  'DESCRIPTOR' : _APEXINFO,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.ApexInfo)
  })
_sym_db.RegisterMessage(ApexInfo)

ApexMetadata = _reflection.GeneratedProtocolMessageType('ApexMetadata', (_message.Message,), {
  'DESCRIPTOR' : _APEXMETADATA,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.ApexMetadata)
  })
_sym_db.RegisterMessage(ApexMetadata)

DeltaArchiveManifest = _reflection.GeneratedProtocolMessageType('DeltaArchiveManifest', (_message.Message,), {
  'DESCRIPTOR' : _DELTAARCHIVEMANIFEST,
  '__module__' : 'update_metadata_pb2'
  # @@protoc_insertion_point(class_scope:chromeos_update_engine.DeltaArchiveManifest)
  })
_sym_db.RegisterMessage(DeltaArchiveManifest)


DESCRIPTOR._options = None
_SIGNATURES_SIGNATURE.fields_by_name['version']._options = None
_IMAGEINFO.fields_by_name['board']._options = None
_IMAGEINFO.fields_by_name['key']._options = None
_IMAGEINFO.fields_by_name['channel']._options = None
_IMAGEINFO.fields_by_name['version']._options = None
_IMAGEINFO.fields_by_name['build_channel']._options = None
_IMAGEINFO.fields_by_name['build_version']._options = None
_INSTALLOPERATION_TYPE.values_by_name["MOVE"]._options = None
_INSTALLOPERATION_TYPE.values_by_name["BSDIFF"]._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['install_operations']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['kernel_install_operations']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['old_kernel_info']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['new_kernel_info']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['old_rootfs_info']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['new_rootfs_info']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['old_image_info']._options = None
_DELTAARCHIVEMANIFEST.fields_by_name['new_image_info']._options = None
# @@protoc_insertion_point(module_scope)



"""Miscellaneous functions used by the tool."""
import re
from pathlib import Path
from string import Template
from typing import Union


def extract_codename(updater_script) -> str:
    r"""Extract device codename form updater-script.

    Regex pattern explanation:
    - (?:/[\w\d_-]+/([\w\d]+):\d) matches codename in update fingerprint like
    Xiaomi/cmi_global/cmi:11/RKQ1.200710.002/V12.1.2.0.RJAMIXM:user/release-keys
    - (?:\(\"ro\.product\.device\"\) == \"([\w\d]+)\") matches codename
     in getprop("ro.product.device") statements
    - (?:get_device_compatible\(\"([\w\d]+)\"\)) matches get_device_compatible() statements
    :param updater_script: updater-script file as string
    :return: extracted codename if found or 'codename' if not found.
    """
    pattern = re.compile(
        r"(?:/[\w\d_-]+/([\w\d]+):\d)|"
        r"(?:\(\"ro\.product\.device\"\) == \"([\w\d]+)\")|"
        r"(?:get_device_compatible\(\"([\w\d]+)\"\))"
    )
    match = pattern.search(updater_script)
    if match:
        codename = [i for i in match.groups() if i is not None]
        return codename[0]
    return "codename"


def cleanup_codename(codename: str) -> str:
    """
    Remove region names from codename and android one's "SPROUT"
    :param codename: codename from miui zip
    :return: clean codename
    """
    if "SPROUT" in codename:
        codename = codename.replace("SPROUT", "")
    if codename.endswith("PRE"):
        codename = codename.replace("PRE", "")
    if "EEAGlobal" in codename:
        return codename.replace("EEAGlobal", "")
    if "IDGlobal" in codename and codename not in ["CUPIDGlobal"]:
        return codename.replace("IDGlobal", "")
    if "INGlobal" in codename and codename not in ["CHOPINGlobal"]:
        return codename.replace("INGlobal", "")
    if "RUGlobal" in codename:
        return codename.replace("RUGlobal", "")
    if "TRGlobal" in codename:
        return codename.replace("TRGlobal", "")
    if "TWGlobal" in codename:
        return codename.replace("TWGlobal", "")
    if "Global" in codename:
        return codename.replace("Global", "")
    return codename


class ScriptTemplate(Template):
    delimiter = "[-]"


def write_text_to_file(file: Union[str, Path], text: str):
    Path(file).write_bytes(text.encode("utf-8"))



"""Local Zip Extractor is a concrete class that implements BaseExtractor methods \
that deals with local zip files using zipfile standard library."""

from pathlib import Path
from typing import List, Union
from zipfile import ZipFile

from remotezip import RemoteZip
from requests import head

from xiaomi_flashable_firmware_creator.extractors.handlers.android_one_zip import (
    AndroidOneZip,
)
from xiaomi_flashable_firmware_creator.extractors.handlers.standard_zip import (
    StandardZip,
)


class ZipExtractor:
    """ZipExtractor provides methods for dealing with local and remote zip files."""

    zip_file_path: Union[Path, str]
    zip_file: str
    zip_url: str
    handler: Union[StandardZip, AndroidOneZip]

    def __init__(self, zip_file, tmp_dir):
        """
        Initialize LocalZipExtractor.

        :param zip_file: a path object or a string to a zip that contains a full recovery ROM.
        :param tmp_dir: output directory to place the extracted zip in.
        """
        self.zip_url = (
            zip_file if "http" in zip_file or "ota.d.miui.com" in zip_file else ""
        )
        self.zip_file_path = (
            Path(zip_file) if not self.zip_url and isinstance(zip_file, str) else ""
        )
        self.files = []
        self._extractor = (
            RemoteZip(self.zip_url) if self.zip_url else ZipFile(self.zip_file_path)
        )
        self.handler = (
            AndroidOneZip(self.zip_file_path, tmp_dir, self._extractor)
            if "payload.bin" in str(self._extractor.namelist())
            else StandardZip(self.zip_file_path, tmp_dir, self._extractor)
        )

    def exists(self) -> bool:
        """
        Check if the local zip file exists.

        :return: True if zip file exists, False otherwise.
        """
        return (
            self.zip_file_path.exists() if self.zip_file_path else head(self.zip_url).ok
        )

    def get_files_list(self):
        """
        Get contents of the local zip file as list and store it in files attribute.

        :return:
        """
        self.files = self._extractor.namelist()

    def get_file_name(self) -> str:
        """
        Get input zip file name and return it as a string.

        :return: a string of the input zip file name.
        """
        return (
            self.zip_file_path.name
            if self.zip_file_path
            else self.zip_url.split("/")[-1]
        )

    def prepare(self):
        if isinstance(self.handler, AndroidOneZip):
            self.files = self.handler.prepare()

    def extract(self, files_to_extract: List[str]):
        """
        Extract a list of files from the zip file

        :param files_to_extract: a list of files to extract
        :return:
        """
        self.handler.extract(files_to_extract)

    def close(self):
        """
        Close the zip file.

        :return:
        """
        self._extractor.close()



from xiaomi_flashable_firmware_creator.extractors.handlers.base_handler import (
    BaseHandler,
)


class StandardZip(BaseHandler):
    def __init__(self, zip_file_path, tmp_dir, extractor):
        super().__init__(zip_file_path, tmp_dir, extractor)



from pathlib import Path
from typing import List

from xiaomi_flashable_firmware_creator.extractors.handlers.base_handler import (
    BaseHandler,
)
from xiaomi_flashable_firmware_creator.extractors.ota_payload_extractor.extract_android_ota_payload import (
    Payload,
    parse_payload,
)

# from google.protobuf.pyext._message import RepeatedCompositeContainer


class AndroidOneZip(BaseHandler):
    payload: Payload

    # partitions: Dict[str, RepeatedCompositeContainer]

    def __init__(self, zip_file_path, tmp_dir, extractor):
        super().__init__(zip_file_path, tmp_dir, extractor)
        self.payload_file = None
        self.all_partitions = set()
        self.files = []

    def prepare(self) -> List[str]:
        """
        Extract payload from zip in order to get partition names
        :return: a list of partition names strings
        """
        self.payload_file = open(
            self.extractor.extract("payload.bin", self._tmp_dir), "rb"
        )
        self.payload = Payload(self.payload_file)
        self.payload.Init()
        self.partitions = {
            i.partition_name: i for i in self.payload.manifest.partitions
        }
        self.all_partitions = set(self.partitions.keys())
        self.files = sorted([f"firmware-update/{i}.img" for i in self.all_partitions])
        return self.files

    def extract(self, files_to_extract: List[str]):
        """
        Extract partitions from payload file
        :param files_to_extract: a list of files to extract
        :return:
        """
        Path(self._tmp_dir / "firmware-update").mkdir(parents=True, exist_ok=True)
        files_to_extract: set = set(self.files).intersection(set(files_to_extract))
        for file in files_to_extract:
            # partition: RepeatedCompositeContainer = self.partitions.get(file.split('/')[-1].split('.')[0])
            partition = self.partitions.get(file.split("/")[-1].split(".")[0])
            with open(
                Path(self._tmp_dir / f"firmware-update/{partition.partition_name}.img"),
                "wb",
            ) as out_f:
                parse_payload(self.payload, partition, out_f)
        self.payload_file.close()
        Path(self._tmp_dir / "payload.bin").unlink()



from typing import List, Union
from zipfile import ZipFile

from remotezip import RemoteZip


class BaseHandler:
    extractor: Union[ZipFile, RemoteZip]

    def __init__(self, zip_file_path, tmp_dir, extractor):
        self.zip_file_path = zip_file_path
        self._tmp_dir = tmp_dir
        self.extractor = extractor

    def extract(self, files_to_extract: List[str]):
        self.extractor.extractall(path=self._tmp_dir, members=files_to_extract)



"""This module provides ProcessTypes and ZipTypes classes that handle some \
type checks of the tool process."""

from enum import Enum, auto


class ProcessTypes(Enum):
    """
    A process type enum that defines the type of the current run process.

    Values are firmware, non_arb_firmware, firmware_less and vendor.
    """

    firmware = auto()
    non_arb_firmware = auto()
    firmware_less = auto()
    vendor = auto()


class ZipTypes(Enum):
    """
    An enum that represents supported zip ROM types.

    Values are qcom for Qualcomm devices' ROMs and mtk for Mediatek devices' ROMs.
    """

    qcom = auto()
    mtk = auto()



"""
Xiaomi Flashable Firmware Creator class.

Xiaomi Flashable Firmware Creator class provides an easy way to
   create a flashable firmware zip file from any supported ROM.
Using extract, generate_updater_script, and make_zip you can create
   flashable files according to the provided configuration
   ('firmware', 'nonarb', 'firmwareless', 'vendor) .
"""

# pylint: disable=too-many-instance-attributes
import re
from datetime import datetime
from pathlib import Path
from shutil import copy2, make_archive, rmtree
from socket import gethostname
from typing import List

from xiaomi_flashable_firmware_creator import work_dir
from xiaomi_flashable_firmware_creator.extractors.handlers.android_one_zip import (
    AndroidOneZip,
)
from xiaomi_flashable_firmware_creator.extractors.zip_extractor import ZipExtractor
from xiaomi_flashable_firmware_creator.helpers.misc import (
    ScriptTemplate,
    cleanup_codename,
    extract_codename,
    write_text_to_file,
)
from xiaomi_flashable_firmware_creator.types import ProcessTypes, ZipTypes


class FlashableFirmwareCreator:
    """FlashableFirmwareCreator provides methods for creating \
     flashable firmware files from Xiaomi supported ROMs."""

    extractor: ZipExtractor
    input_file: str
    _tmp_dir: Path
    _flashing_script_dir: Path
    host: str
    datetime: datetime
    _extract_mode: str
    extract_mode: ProcessTypes
    type: ZipTypes
    update_script: str
    is_android_one: bool

    def __init__(self, input_file, _extract_mode, out_dir=""):
        """
        Initialize FlashableFirmwareCreator.

        :param input_file: zip file to extract from.
         It can be a local path or a remote direct url.
        :param _extract_mode: Which mode should the tool use.
         This must be one of "firmware", "nonarb", "firmwareless" or "vendor".
        :param out_dir: The output directory to store the extracted file in.
        """
        self.input_file = input_file
        _tmp_sub_dir = f"{Path(input_file).stem.split('_')[-2]}_"  # set tmp subdirectory to ROM's hash
        self._tmp_dir = (
            Path(out_dir) / "tmp" / _tmp_sub_dir
            if out_dir
            else work_dir / "tmp" / _tmp_sub_dir
        )
        self._out_dir = self._tmp_dir.parent.parent.absolute()
        self._flashing_script_dir = (
            self._tmp_dir.absolute() / "META-INF/com/google/android"
        )
        self.host = gethostname()
        self.datetime = datetime.now()
        self.extract_mode = self.get_extract_mode(_extract_mode)
        self.update_script = ""
        self.is_android_one = False
        self.firmware_excluded_files = [
            "dtbo",
            "logo",
            "splash",
            "vbmeta",
            "boot.img",
            "recovery.img",
            "system",
            "vendor",
            "product",
            "odm",
            "exaid",
            "mi_ext",
            "dynamic_partitions_op_list",
            "metadata.pb",
        ]
        self.vendor_excluded_files = [
            "vbmeta",
            "system",
            "product.",
            "odm.",
            "metadata.pb",
        ]
        self.extractor = ZipExtractor(self.input_file, self._tmp_dir)
        self.init()

    def init(self):
        """
        Do some initial checks and housekeeping.

        - This method checks if the temporary directory exists
         and removes it then creates a new one.
        - Creates updater-script temporary path.
        - Checks if the input zip file exists and
         raise a FileNotFoundError exception if not.
        - Open the zip and get contents list.
        - Checks if the zip file is a valid ROM. If true get the ROM type.
         Otherwise, remove the temporary directory and raise a RuntimeError.
        """
        if self._tmp_dir.exists():
            rmtree(self._tmp_dir)
        self._tmp_dir.mkdir(parents=True, exist_ok=True)
        self._flashing_script_dir.mkdir(parents=True, exist_ok=True)
        if not self.extractor.exists():
            raise FileNotFoundError(f"input file {self.input_file} does not exist!")
        self.extractor.get_files_list()
        if not self.is_valid_rom():
            rmtree(self._tmp_dir)
            raise RuntimeError(f"{self.input_file} is not a valid ROM file. Exiting..")

    @staticmethod
    def get_extract_mode(extract_mode) -> ProcessTypes:
        """
        Get the extract mode enum according to extract_mode string value.

        :param extract_mode: "firmware", "nonarb", "firmwareless" or "vendor"
        :return: ProcessTypes enum
        """
        modes = {
            "firmware": ProcessTypes.firmware,
            "nonarb": ProcessTypes.non_arb_firmware,
            "firmwareless": ProcessTypes.firmware_less,
            "vendor": ProcessTypes.vendor,
        }
        try:
            return modes[extract_mode]
        except KeyError as err:
            print("Unknown process!")
            raise err

    def is_valid_rom(self) -> bool:
        """
        Check if the zip file is valid ROM.

        :return: True if update-binary or updater-script or payload.bin is present
         in contents list, False otherwise.
        """
        return (
            "META-INF/com/google/android/update-binary" in self.extractor.files
            or "META-INF/com/google/android/updater-script" in self.extractor.files
            or "payload.bin" in self.extractor.files
        )

    def get_rom_type(self):
        """
        Get the type of the input ROM or raise a RuntimeError.

        :return: An enum of ROM type. Either qcom or mtk.
        """
        files = str(self.extractor.files)
        if "lk.img" in files or "preloader.img" in files:
            self.type = ZipTypes.mtk
        elif (
            "firmware-update" in files
            or "rpm" in files
            or "tz" in files
            or "keymaster" in files
        ):
            self.type = ZipTypes.qcom
        else:
            raise RuntimeError("Can't detect rom type. It's not qcom or mtk!'")

    def get_files_list(self) -> List[str]:
        """
        Get files to extract list according to the extract mode Enum.

        :return: a list of file names string.
        """
        # TODO: match fastboot and android one img files
        #  when no firmware-update directory is found
        if self.extract_mode is ProcessTypes.firmware:
            return (
                [
                    i
                    for i in self.extractor.files
                    if (
                        i.startswith("META-INF/")
                        and (
                            i.endswith("updater-script") or i.endswith("update-binary")
                        )
                    )
                    or all(n not in i for n in self.firmware_excluded_files)
                ]
                if self.type is ZipTypes.qcom
                else [
                    n
                    for n in self.extractor.files
                    if all(
                        file not in n
                        for file in self.firmware_excluded_files + ["file_contexts"]
                    )
                ]
            )
        if self.extract_mode is ProcessTypes.non_arb_firmware:
            return [
                n
                for n in self.extractor.files
                if "dspso.bin" in n
                or n.startswith("firmware-update/BTFM.bin")
                or n.startswith("firmware-update/NON-HLOS.bin")
                or n.startswith("META-INF/")
                and (n.endswith("updater-script") or n.endswith("update-binary"))
            ]
        if self.extract_mode is ProcessTypes.firmware_less:
            return (
                [
                    n
                    for n in self.extractor.files
                    if not n.startswith("firmware-update/")
                ]
                if self.type is ZipTypes.qcom
                else []
            )
        if self.extract_mode is ProcessTypes.vendor:
            return [
                i
                for i in self.extractor.files
                if all(n not in i for n in self.vendor_excluded_files)
            ]
        return []  # Will never happen

    def get_updater_script_lines(self, invalid_files: set[str]) -> str:
        """
        Get selective lines from the update-script according to process type enum value.

        :return: a string of the updater-script lines
        """
        original_updater_script = Path(self._flashing_script_dir / "updater-script")
        if not original_updater_script.exists():
            raise FileNotFoundError("updater-script not found!")
        original_updater_script = original_updater_script.read_text().splitlines()
        lines = []
        if self.extract_mode is ProcessTypes.firmware:
            lines = (
                [
                    line
                    for line in original_updater_script
                    if ("getprop" in line and "ro.build.date.utc" not in line)
                    or "Target" in line
                    or (
                        "firmware-update" in line
                        and line.split("/")[1].split('"')[0] not in invalid_files
                    )
                    and (
                        "ro.product" in line
                        or all(
                            file not in line for file in self.firmware_excluded_files
                        )
                    )
                ]
                if self.type is ZipTypes.qcom
                else [
                    line
                    for line in original_updater_script
                    if "ro.product" in line
                    or all(file not in line for file in self.firmware_excluded_files)
                ]
            )
        elif self.extract_mode is ProcessTypes.non_arb_firmware:
            lines = [
                line
                for line in original_updater_script
                if ("getprop" in line and "ro.build.date.utc" not in line)
                or "Target" in line
                or "modem" in line
                or "bluetooth" in line
                or "dsp" in line
            ]
        elif self.extract_mode is ProcessTypes.firmware_less:
            lines = (
                [
                    line
                    for line in original_updater_script
                    if ("getprop" in line and "ro.build.date.utc" not in line)
                    or "Target" in line
                    or "boot.img" in line
                    or "system" in line
                    or "vendor" in line
                    or "product" in line
                ]
                if self.type is ZipTypes.qcom
                else []
            )
        elif self.extract_mode is ProcessTypes.vendor:
            lines = (
                [
                    line
                    for line in original_updater_script
                    if ("getprop" in line and "ro.build.date.utc" not in line)
                    or "Target" in line
                    or "dynamic_partitions_op_list" in line
                    or (
                        "firmware-update" in line
                        and line.split("/")[1].split('"')[0] not in invalid_files
                    )
                    and "vbmeta" not in line
                    or "vendor" in line
                ]
                if self.type is ZipTypes.qcom
                else [
                    line
                    for line in original_updater_script
                    if "system" not in line and "boot.img" not in line
                ]
            )
        return "\n".join(lines)

    def generate_updater_script(self, invalid_files: set[str]):
        """
        Generate the new zip updater-script and write it to the temporary directory.

        :return:
        """
        template = ScriptTemplate(
            Path(
                Path(__file__).parent / "templates/recovery_updater_script"
            ).read_text()
        )
        lines = self.get_updater_script_lines(invalid_files)
        if not lines:
            raise RuntimeError("Could not extract lines from updater-script!")
        if self.extract_mode is ProcessTypes.firmware:
            process = "Normal Firmware"
        elif self.extract_mode is ProcessTypes.non_arb_firmware:
            process = "non-ARB firmware"
        elif self.extract_mode is ProcessTypes.firmware_less:
            process = "firmware-less ROM"
        elif self.extract_mode is ProcessTypes.vendor:
            process = "firmware+vendor"
        else:
            process = "Unknown"  # This should never happen

        updater_script = template.substitute(
            datetime=self.datetime, host=self.host, process=process, lines=lines
        )
        # correct some updater-script lines that exist in old devices' file
        if "/firmware/image/sec.dat" in updater_script:
            updater_script = updater_script.replace(
                "/firmware/image/sec.dat", "/dev/block/bootdevice/by-name/sec"
            )
        if "/firmware/image/splash.img" in updater_script:
            updater_script = updater_script.replace(
                "/firmware/image/splash.img", "/dev/block/bootdevice/by-name/splash"
            )
        self.update_script = updater_script
        write_text_to_file(
            f"{str(self._flashing_script_dir)}/updater-script", updater_script
        )
        # Use modified dynamic_partitions_op_list with resize vendor line only
        if (
            self.extract_mode is ProcessTypes.vendor
            and "dynamic_partitions_op_list" in updater_script
        ):
            original_dynamic_partitions_list = Path(
                self._tmp_dir / "dynamic_partitions_op_list"
            ).read_text()
            vendor_resize = re.search(
                r"(resize vendor .*$)", original_dynamic_partitions_list, re.M
            )
            if vendor_resize:
                write_text_to_file(
                    f"{str(self._tmp_dir)}/dynamic_partitions_op_list",
                    vendor_resize.group(1),
                )

    # def generate_update_binary(self):
    #     """
    #     Generate the new zip update-binary and write it to the temporary directory.
    #
    #     :return:
    #     """
    #     template = ScriptTemplate(Path(
    #         Path(__file__).parent / 'templates/recovery_update-binary').read_text())
    #     update_binary_text = template.substitute(datetime=self.datetime, host=self.host)
    #     update_binary = Path(f"{str(self._flashing_script_dir)}/update-binary")
    #     update_binary.write_text(update_binary_text)
    #     update_binary.chmod(775)

    def generate_ab_updater_script(self, invalid_files: set[str]):
        script_template = ScriptTemplate(
            Path(
                Path(__file__).parent / "templates/recovery_ab_updater_script"
            ).read_text()
        )
        flashing_template = ScriptTemplate(
            Path(Path(__file__).parent / "templates/partition_flashing").read_text()
        )
        lines = [
            flashing_template.substitute(partition=file.split("/")[-1].split(".")[0])
            for file in self.get_files_list()
            if file.startswith("firmware-update/") and file not in invalid_files
        ]
        updater_script = script_template.substitute(
            datetime=self.datetime,
            host=self.host,
            zip_name=self.extractor.get_file_name(),
            lines="\n".join(lines),
        )
        write_text_to_file(
            f"{str(self._flashing_script_dir)}/updater-script", updater_script
        )

    def generate_flashing_script(self, invalid_files: set[str]):
        """
        Generate update_script or update-binary
        :return:
        """
        if self.is_android_one is True:
            self.generate_ab_updater_script(invalid_files)
            copy2(
                Path(Path(__file__).parent / "binaries/update-binary"),
                f"{str(self._flashing_script_dir)}/update-binary",
            )
        else:
            self.generate_updater_script(invalid_files)

    def make_zip(self) -> str:
        """
        Compress the temporary directory into a zip archive in output directory.

        Also, name it according to process parameters.
        :return:
        """
        out = Path(f"{self._out_dir}/result.zip")
        make_archive(str(out.with_suffix("").absolute()), "zip", self._tmp_dir)
        if not out.exists():
            raise RuntimeError("Could not create result zip file!")
        if not self.is_android_one:
            codename = extract_codename(self.update_script)
        else:
            codename = cleanup_codename(
                self.extractor.get_file_name().split("_")[1]
            ).lower()
        zip_prefix = ""
        if self.extract_mode is ProcessTypes.firmware:
            zip_prefix = "fw"
        elif self.extract_mode is ProcessTypes.non_arb_firmware:
            zip_prefix = "fw-non-arb"
        elif self.extract_mode is ProcessTypes.firmware_less:
            zip_prefix = "fw-less"
        elif self.extract_mode is ProcessTypes.vendor:
            zip_prefix = "fw-vendor"
        else:
            pass  # This should never happen
        zip_name = (
            f"{self._out_dir}/{zip_prefix}_{codename}_{self.extractor.get_file_name()}"
        )
        out.rename(zip_name)
        return zip_name

    def extract(self) -> tuple[set[str], set[str]]:
        """
        Invoke the extract method of extractor object.

        :return: a tuple of a set of valid files to extract and a set of zero length invalid files.
        """
        if hasattr(self.extractor, "handler") and isinstance(
            self.extractor.handler, AndroidOneZip
        ):
            self.is_android_one = True
            self.extractor.prepare()
        self.get_rom_type()
        files_to_extract = self.get_files_list()
        if not files_to_extract or (
            len(files_to_extract) == 2
            and "updater-script" in str(files_to_extract)
            and "update-binary" in str(files_to_extract)
        ):
            self.cleanup()
            self.close()
            raise RuntimeError("Nothing found to extract!")
        self.extractor.extract(files_to_extract)
        # Filter out zero length invalid files
        invalid_files = set()
        firmware_update_dir = Path(self._tmp_dir / "firmware-update")
        if not firmware_update_dir.exists():
            return set(files_to_extract), invalid_files
        for file in firmware_update_dir.iterdir():
            if file.is_file() and file.stat().st_size == 0:
                file.unlink(missing_ok=True)
                invalid_files.add(file.name)
        return set(files_to_extract).difference(invalid_files), invalid_files

    def cleanup(self):
        """
        Cleanup temporary directory and files.

        :return:
        """
        rmtree(self._tmp_dir)

    def close(self):
        """
        Invoke close method of the extractor object.

        :return:
        """
        self.extractor.close()

    def auto(self) -> str:
        """
        Autopilot method for doing everything with a single call.

        This is useful for using this class from a command line interface
         or another application.
        :return: output zip file name as a string.
        """
        _, invalid_files = self.extract()
        print("Generating flashing script...")
        self.generate_flashing_script(invalid_files)
        print("Creating new zip file..")
        new_zip = self.make_zip()
        self.cleanup()
        self.close()
        return new_zip
